"""Outset Python Package"""
from __future__ import annotations

__version__ = "0.0.0""-rc94-post1"

from .CalcBoundsIQR_ import CalcBoundsIQR
from .draw_marquee_ import draw_marquee
from .inset_outsets_ import inset_outsets
from .layout_corner_insets_ import layout_corner_insets
from .MarkArrow_ import mark_arrow, MarkArrow
from .MarkInlaidAsterisk_ import mark_inlaid_asterisk, MarkInlaidAsterisk
from .MarkMagnifyingGlass_ import mark_magnifying_glass, MarkMagnifyingGlass
from .MarkAlphabeticalBadges_ import MarkAlphabeticalBadges
from .MarkNumericalBadges_ import MarkNumericalBadges
from .MarkRomanBadges_ import MarkRomanBadges
from .OutsetGrid_ import OutsetGrid
from .marqueeplot_ import marqueeplot
from .rescale_clip_outliers_ import rescale_clip_outliers
from .stub_all_clipped_values_ import stub_all_clipped_values
from .stub_edge_mark_ import stub_edge_mark

__all__ = [
    "CalcBoundsIQR",
    "draw_marquee",
    "inset_outsets",
    "layout_corner_insets",
    "mark_arrow",
    "mark_inlaid_asterisk",
    "mark_magnifying_glass",
    "MarkAlphabeticalBadges",
    "MarkArrow",
    "MarkInlaidAsterisk",
    "MarkMagnifyingGlass",
    "MarkNumericalBadges",
    "MarkRomanBadges",
    "marqueeplot",
    "OutsetGrid",
    "rescale_clip_outliers",
    "stub_all_clipped_values",
    "stub_edge_mark",
]



import typing

import matplotlib.pyplot as plt
from matplotlib import axes as mpl_axes

_color_t = typing.Union[typing.Tuple, str]


def mark_inlaid_asterisk(
    x: typing.Union[float, typing.Sequence[float]],
    y: typing.Union[float, typing.Sequence[float]],
    ax: typing.Optional[mpl_axes.Axes] = None,
    *,
    asterisk_edgewidth: float = 1,
    color: str = "black",
    color_accent: str = "white",
    color_asterisk_edge: _color_t = ("color", 0.7),
    color_asterisk_face: _color_t = ("color", 0.7),
    color_badge: _color_t = ("color", 0.3),
    color_underlay: _color_t = "color_accent",
    marker: typing.Tuple = (6, 2, 0),  # 6 points, rotated 0 degrees
    marker_badge: str = "o",
    marker_underlay: str = "o",
    linecolor: str = "none",
    markersize: float = 22,
    scale_asterisk: float = 0.3,
    scale_badge: float = 0.8,
    **kwargs,
) -> None:
    """Draw asterisk badge marker(s) at specified location(s) on a matplotlib p
    lot.

    This function stacks matplotlib markers to render an asterisk overlaid onto
    a circular badge with a slightly larger circular underlay.

    Parameters
    ----------
    x : float or sequence of floats
        The x-coordinate(s) where the asterisk badge(s) will be placed.
    y : float or sequence of floats
        The y-coordinate(s) where the asterisk badge(s) will be placed.
    ax : mpl_axes.Axes, optional
        The axes object on which to draw the markers. If None, plt.gca() will
        be used.
    asterisk_edgewidth : float, default 1
        The edge width of the asterisk marker.
    color : str, default "black"
        The primary color for the glyph.
    color_asterisk_edge : Union[str, tuple], default ("color", 0.7)
        The edge color of the asterisk marker.

        If "color", primary color will be substituted.
    color_asterisk_face : Union[str, tuple], default ("color", 0.7)
        The face color of the asterisk marker.

        If "color", primary color will be substituted.
    color_badge : Union[str, tuple], default ("color", 0.3)
        The color of the badge.

        If "color", primary color will be substituted.
    color_underlay : Union[str, tuple], default "white"
        The color of the underlay.
    marker : tuple, default (6, 2, 0)
        The marker style for the asterisk. Default is 6 points, rotated 0
        degrees.
    marker_badge : str, default "o"
        The marker style for the badge.
    marker_underlay : str, default "o"
        The marker style for the underlay.
    linecolor : str, default "none"
        The color of the line connecting markers, if any.
    markersize : float, default 22
        Size for glyph's largest marker element.
    scale_asterisk : float, default 0.3
        The scaling factor for the asterisk size.
    scale_badge : float, default 0.8
        The scaling factor for the badge size.
    **kwargs : dict, optional
        Additional keyword arguments forward to matplotlib `plot`.

    Returns
    -------
    None

    See Also
    --------
    MarkInlaidAsterisk: Functor interface for `mark_inlaid_asterisk`.
    """
    if ax is None:
        ax = plt.gca()

    # substitute "color" with color kwarg
    def substitute_color(query: _color_t) -> _color_t:
        if query == "color":
            return color
        elif query == "color_accent":
            return color_accent
        elif isinstance(query, tuple):
            if not query:
                raise ValueError("color tuple must have at least one element")
            first, *rest = query
            if first == "color":
                return (color, *rest)
            elif first == "color_accent":
                return (color_accent, *rest)
            else:
                return query
        else:
            return query

    colors = {
        "color_asterisk_edge": substitute_color(color_asterisk_edge),
        "color_asterisk_face": substitute_color(color_asterisk_face),
        "color_badge": substitute_color(color_badge),
        "color_underlay": substitute_color(color_underlay),
    }

    # Base underlay
    ax.plot(
        x,
        y,
        color=linecolor,
        marker=marker_underlay,
        markeredgecolor="none",
        markerfacecolor=colors["color_underlay"],
        markersize=markersize,
        **kwargs,
    )

    # Badge
    ax.plot(
        x,
        y,
        color="none",
        marker=marker_badge,
        markeredgecolor="none",
        markerfacecolor=colors["color_badge"],
        markersize=markersize * scale_badge,
        **kwargs,
    )

    # Asterisk
    ax.plot(
        x,
        y,
        color="none",
        marker=marker,
        markeredgewidth=asterisk_edgewidth,
        markeredgecolor=colors["color_asterisk_edge"],
        markerfacecolor=colors["color_asterisk_face"],
        markersize=markersize * scale_asterisk,
        **kwargs,
    )


class MarkInlaidAsterisk:
    """Functor interface for `mark_inlaid_asterisk`."""

    def __call__(self, *args, **kwargs):
        """Forwards to `mark_inlaid_asterisk`."""
        return mark_inlaid_asterisk(*args, **kwargs)



import typing

import matplotlib.pyplot as plt
from matplotlib import axes as mpl_axes
from matplotlib import markers as mpl_markers

from ._auxlib.rotate_marker_ import rotate_marker


def mark_arrow(
    x: typing.Union[float, typing.Sequence[float]],
    y: typing.Union[float, typing.Sequence[float]],
    ax: typing.Optional[mpl_axes.Axes] = None,
    *,
    alpha: float = 1.0,
    color: str = "black",
    color_accent: str = "white",
    linecolor: str = "none",
    markersize: float = 15,
    rotate_angle: float = 20,
    **kwargs,
) -> None:
    """Draw arrow marker(s) at specified location(s) on a matplotlib plot.

    This function creates an arrow glyph by combining and rotating matplotlib
    markers.

    Parameters
    ----------
    x : float or sequence of floats
        The x-coordinate(s) where the arrow(s) will be placed.
    y : float or sequence of floats
        The y-coordinate(s) where the arrow(s) will be placed.
    ax : mpl_axes.Axes, optional
        The axes object on which to draw the markers. If None, `plt.gca()` will
        be used.
    alpha : float, default 1.0
        The transparency level of the markers.
    color : str, default "black"
        The primary color for the glyph components.
    color_accent : str, default "white"
        The color used for the accent layer of the glyph, enhancing visibility.
    linecolor : str, default "none"
        Color for connecting lines between markers, if any.
    markersize : float, default 15
        Size for glyph's largest marker element.
    rotate_angle : float, default 20
        The angle (in degrees) to rotate the arrow part of the glyph.
    **kwargs : dict, optional
        Additional keyword arguments forward to matplotlib `plot`.

    Returns
    -------
    None
    """
    if ax is None:
        ax = plt.gca()

    # rotated tickdown marker for handle
    # adapted from https://stackoverflow.com/a/49662573
    head_marker = rotate_marker(10, rotate_angle)  # CARETUP
    stem_marker = rotate_marker(3, rotate_angle)  # TICKDOWN

    # Scaling factors for different elements relative to markersize
    overlay_scale = 0.8
    stem_scale = 0.6

    # Underlay stem
    ax.plot(
        x,
        y,
        alpha=alpha,
        color=linecolor,
        marker=stem_marker,
        markeredgecolor=color_accent,
        markeredgewidth=markersize * 0.75,
        markersize=markersize * stem_scale,
        **kwargs,
    )

    # Draw stem
    ax.plot(
        x,
        y,
        alpha=0.8 * alpha,
        color="none",  # line rendering is handled above
        marker=stem_marker,
        markeredgecolor=color,
        markeredgewidth=markersize * 0.5,
        markerfacecolor=color,
        markersize=markersize * overlay_scale * stem_scale,
        **kwargs,
    )

    # Underlay arrowhead
    ax.plot(
        x,
        y,
        alpha=alpha,
        color="none",  # line rendering is handled above
        marker=head_marker,
        markeredgecolor=color_accent,
        markeredgewidth=markersize / 12,
        markerfacecolor=color_accent,
        markersize=markersize,
        **kwargs,
    )

    # draw arrowhead
    ax.plot(
        x,
        y,
        alpha=0.8 * alpha,
        color="none",  # line rendering is handled above
        marker=head_marker,
        markeredgecolor="none",
        markerfacecolor=color,
        markersize=markersize * overlay_scale,
        **kwargs,
    )



import copy
import typing
import warnings

import frozendict
from matplotlib import axes as mpl_axes
from matplotlib import pyplot as plt
import opytional as opyt
import pandas as pd
import seaborn as sns

from ._auxlib.calc_aspect_ import calc_aspect
from ._auxlib.equalize_aspect_ import equalize_aspect
from ._auxlib.set_aspect_ import set_aspect
from .MarkNumericalBadges_ import MarkNumericalBadges
from .marqueeplot_ import marqueeplot


class OutsetGrid(sns.axisgrid.FacetGrid):
    """Facilitates co-display of zoomed-in axis regions transplanted across a
    subplot grid.

    Corresponding regions in the original ("source") plot and the zoomed-in,
    "outset" regions are marked with corresponding "marquee" annotations. The
    OutsetGrid may be configured to include the source plot on the first
    subplot axis, or to only display outset regions on subplot axes.

    Marquee annotation creation must be dispatched manually through
    `marqueeplot`, `marqueeplot_outset`, and/or `marqueeplot_source`. This
    mechanism allows for end-user control over plot sequencing (i.e., layering
    order) and for adjustment of axis sizing prior to marquee rendering.
    (Marquee layout dimensions are sensitive to axis rescaling.)

    Inherits from seaborn's FacetGrid, so FacetGrad API components are
    available, see <https://seaborn.pydata.org/generated/seaborn.FacetGrid.html>.

    Attributes
    ----------
    source_axes : Optional[mpl_axes.Axes]
        The axes object for the source plot, if present.
    outset_axes : Sequence[mpl_axes.Axes]
        The axes objects for the outset plots.
    """

    __data: pd.DataFrame
    _marqueeplot_outset: typing.Callable
    _marqueeplot_source: typing.Callable

    source_axes: typing.Optional[mpl_axes.Axes]
    outset_axes: typing.Sequence[mpl_axes.Axes]

    def tight_layout(self) -> None:
        self.figure.tight_layout()

    def _finalize_grid(self, axlabels) -> None:
        """Finalize the annotations and layout."""
        self.tight_layout()

    def __init__(
        self: "OutsetGrid",
        data: typing.Union[
            pd.DataFrame,
            typing.Sequence[typing.Tuple[float, float, float, float]],
        ],
        *,
        x: typing.Optional[str] = None,
        y: typing.Optional[str] = None,
        col: typing.Union[str, bool, None] = None,
        col_order: typing.Optional[typing.Sequence[str]] = None,
        col_wrap: typing.Optional[int] = None,
        hue: typing.Union[str, bool, None] = None,
        hue_order: typing.Optional[typing.Sequence[str]] = None,
        outset: typing.Optional[str] = None,
        outset_order: typing.Optional[typing.Sequence[str]] = None,
        color: typing.Optional[str] = None,  # pass to override outset hues
        include_sourceplot: bool = True,
        marqueeplot_kwargs: typing.Dict = frozendict.frozendict(),
        marqueeplot_outset_kwargs: typing.Dict = frozendict.frozendict(),
        marqueeplot_source_kwargs: typing.Dict = frozendict.frozendict(),
        palette: typing.Optional[typing.Sequence] = None,
        zorder: float = 0.0,
        **kwargs,
    ) -> None:
        """Create an OutsetGrid with specified configuration.

        The arguments `data`, `x`, `y`, `col`, `hue`, and `outset` follow
        seaborn-like tidy data API convention. Note that the same value may
        be specified for more than one of `col`, `hue, and `outset`. Marquee
        annotations are created to contain each subset of x, y values with
        identical `col`, `hue`, and `outset` values.

        Marquee frame coordinates may also be specified directly as a sequence
        of four-element tuples `(x0, x1, y0, y1)`. In this case, each frame
        will get its own hue and outset plot.

        Marquee annotation creation must be dispatched manually after
        initialization through `marqueeplot`, `marqueeplot_outset`, and/or
        `marqueeplot_source`.

        Parameters
        ----------
        data : pd.DataFrame or Sequence of Tuple[float, float, float, float]
            A DataFrame containing the data for plotting, or a sequence `(x0,
            x1, y0, y1)` specifying the bounds of outset frames.
        x : Optional[str], default None
            Column name to be used for x-axis values. Not required if data is a sequence of outset frames.
        y : Optional[str], default None
            Column name to be used for y-axis values. Not required if data is a sequence of outset frames.
        col : Union[str, bool, None], default None
            Column name for categorical variable to facet across subaxes.

            If None or True, set to match `outset` or `hue` if provided. If False, no faceting is performed.
        col_order : Optional[Sequence[str]], default None
            The order to arrange the columns in the grid.
        col_wrap : Optional[int], default None
            Number of columns where axes grid should wrap to a new row.
        hue : Union[str, bool, None], default None
            Column name for categorical variable to determine rendered color of
            data's rendered marquee annotations.

            If None or True, set to match `outset` if provided. If False, no faceting is performed.
        hue_order : Optional[Sequence[str]], default None
            The to assign palette colors to `hue` categorical values.

            May contain all or a subset of `data[hue]` values.
        outset : Optional[str], default None
            Column name for categorical variable to segregate data between marquee annotations.
        outset_order : Optional[Sequence[str]], default None
            Order for plotting the outsets.

            May contain all or a subset of `data[outset]` values.
        color : Optional[str], default None
            Color for all outset annotations. Overrides the palette.
        include_sourceplot : bool, default True
            Whether to include the original source plot in the grid.
        marqueeplot_kwargs : Dict, default frozendict()
            Additional marqueeplot keyword arguments  over all plots.
        marqueeplot_outset_kwargs : Dict, default frozendict()
            Additional marqueeplot keyword arguments specific to outset plots.
        marqueeplot_source_kwargs : Dict, default frozendict()
            Additional marqueeplot keyword specific to the source plot, if
            present.
        palette : Optional[Sequence], default None
            Color palette for the outset hue sequence.
        zorder : float, default 0.0
            The z-order for plotting elements.
        **kwargs : dict
            Additional keyword arguments passed to seaborn's FacetGrid.
        """

        if col is None and col_order is not None:
            raise ValueError("col_order must be None if col not specified")
        if hue is None and hue_order is not None:
            raise ValueError("hue_order must be None if hue not specified")
        if outset is None and outset_order is not None:
            raise ValueError(
                "outset_order must be None if outset not specified"
            )

        for a in "frame_inner_pad":
            if a in marqueeplot_outset_kwargs or a in marqueeplot_source_kwargs:
                warnings.warn(
                    f"Specifying {a} independently for only source or "
                    "outset may cause discrepancies in frame placement",
                )

        default_frame_inner_pad, default_frame_outer_pad = 0.2, 0.1

        # spoof data frame if outset frames are specified directly
        if isinstance(data, pd.DataFrame):
            if x is None or y is None:
                raise ValueError(
                    "x and y kwargs must be provided from column names in data",
                )
            if col is None:
                col = outset or hue
                if (
                    col_order is None
                    and outset is not None
                    and outset_order is not None
                ):
                    col_order = outset_order
                elif (
                    col_order is None
                    and outset is None
                    and hue is not None
                    and hue_order is not None
                ):
                    col_order = hue_order

        else:
            if x is not None or y is not None or outset is not None:
                raise ValueError(
                    "x, y, and outset must not be specified if outset frames "
                    "are specified directly",
                )
            if hue not in (None, True, False):
                raise ValueError(
                    "hue must be None or boolean if outset frames are "
                    "specified directly",
                )
            if col not in (None, True, False):
                raise ValueError(
                    "col must be None or boolean if outset frames are "
                    "specified directly",
                )

            if hue is None:
                hue = True
            if col is None:
                col = True

            x, y, outset = "x", "y", "outset"
            data = pd.DataFrame.from_records(
                [
                    {
                        "x": x,
                        "y": y,
                        "outset": i,
                    }
                    for i, (xmin, xmax, ymin, ymax) in enumerate(data)
                    for x, y in [(xmin, ymin), (xmax, ymax)]
                ],
            )
            default_frame_inner_pad = 0

        if not x in data.columns:
            raise ValueError(f"kwarg x={x} must be provided as column in data")
        if not y in data.columns:
            raise ValueError(f"kwarg x={y} must be provided as column in data")
        if outset is not None and not outset in data.columns:
            raise ValueError("if provided, outset must be a column in data")

        if col is True:
            if outset is None:
                raise ValueError("outset must be provided if col is True")
            col = outset
        elif col is False:
            col = None
        if hue is True:
            if outset is None:
                raise ValueError("outset must be provided if hue is True")
            hue = outset
        elif hue is False:
            hue = None

        if col is not None and not col in data.columns:
            raise ValueError("if provided, col must be a column in data")

        if hue is not None and not hue in data.columns:
            raise ValueError("if provided, hue must be a column in data")

        self.__data = data

        if outset is not None and outset_order is None:
            outset_order = sorted(data[outset].unique())

        if col is None:
            assert "_dummy_col" not in data.columns
            col = "_dummy_col"
            data[col] = 0

        if col_order is None:
            col_order = sorted(data[col].unique())

        if include_sourceplot:
            if None in data[col].unique():
                raise ValueError(
                    "Cannot include source plot if col data includes None",
                )
            col_order = [None] + col_order

        # initialize axes
        #######################################################################
        super().__init__(  # initialize parent FacetGrid
            data,
            col=col,
            col_order=col_order,
            col_wrap=col_wrap,
            hue=hue,
            hue_order=hue_order,
            palette=palette,
            **{
                "legend_out": False,
                "sharex": False,
                "sharey": False,
                **kwargs,
            },
        )

        if include_sourceplot:
            self.outset_axes = self.axes.flat[1:]
            self.source_axes = self.axes.flat[0]
        else:
            self.source_axes = None
            self.outset_axes = self.axes.flat[:]

        # draw sourceplot
        #######################################################################
        def marqueeplot_source(self_: "OutsetGrid") -> None:
            if self_.source_axes is None:
                return
            marqueeplot(
                data,
                x=x,
                y=y,
                hue=hue,
                hue_order=hue_order,
                outset=outset,
                outset_order=outset_order,
                ax=self_.source_axes,
                **{
                    "color": color,
                    "palette": palette,
                    "frame_inner_pad": default_frame_inner_pad,
                    "frame_outer_pad": default_frame_outer_pad,
                    "mark_glyph": MarkNumericalBadges,
                    "tight_axlim": False,
                    "zorder": zorder,
                    **copy.deepcopy(marqueeplot_kwargs),  # for mark_glyph
                    **marqueeplot_source_kwargs,
                    "frame_edge_kwargs": {
                        **marqueeplot_kwargs.get("frame_edge_kwargs", {}),
                        **marqueeplot_source_kwargs.get(
                            "frame_edge_kwargs", {}
                        ),
                    },
                    "frame_face_kwargs": {
                        **marqueeplot_kwargs.get("frame_face_kwargs", {}),
                        **marqueeplot_source_kwargs.get(
                            "frame_face_kwargs", {}
                        ),
                    },
                    "leader_edge_kwargs": {
                        **marqueeplot_kwargs.get("leader_edge_kwargs", {}),
                        **marqueeplot_source_kwargs.get(
                            "leader_edge_kwargs", {}
                        ),
                    },
                    "leader_face_kwargs": {
                        **marqueeplot_kwargs.get("leader_face_kwargs", {}),
                        **marqueeplot_source_kwargs.get(
                            "leader_face_kwargs", {}
                        ),
                    },
                    "mark_glyph_kwargs": {
                        "markersize": 16,
                        **marqueeplot_kwargs.get("mark_glyph_kwargs", {}),
                        **marqueeplot_source_kwargs.get(
                            "mark_glyph_kwargs", {}
                        ),
                    },
                },
            )
            self_.source_axes.set_title("")

        self._marqueeplot_source = marqueeplot_source

        # draw outplots
        #######################################################################
        def marqueeplot_outset(self_: "OutsetGrid") -> None:
            # setup
            for d in marqueeplot_kwargs, marqueeplot_outset_kwargs:
                for k, v in d.items():
                    if k == "mark_glyph" and isinstance(v, type):
                        d[k] = v()

            self_.map_dataframe_outset(
                marqueeplot,
                x=x,
                y=y,
                outset=outset,
                outset_order=outset_order,
                **{
                    "color": color,
                    "palette": palette,
                    "frame_inner_pad": default_frame_inner_pad,
                    "frame_outer_pad": default_frame_outer_pad,
                    "leader_stretch": 0.0,
                    "mark_glyph": MarkNumericalBadges(),
                    "tight_axlim": True,
                    "zorder": zorder,
                    **marqueeplot_kwargs,
                    **marqueeplot_outset_kwargs,
                    "frame_edge_kwargs": {
                        **marqueeplot_kwargs.get("frame_edge_kwargs", {}),
                        **marqueeplot_outset_kwargs.get(
                            "frame_edge_kwargs", {}
                        ),
                    },
                    "frame_face_kwargs": {
                        **marqueeplot_kwargs.get("frame_face_kwargs", {}),
                        **marqueeplot_outset_kwargs.get(
                            "frame_face_kwargs", {}
                        ),
                    },
                    "leader_edge_kwargs": {
                        **marqueeplot_kwargs.get("leader_edge_kwargs", {}),
                        **marqueeplot_outset_kwargs.get(
                            "leader_edge_kwargs", {}
                        ),
                    },
                    "leader_face_kwargs": {
                        **marqueeplot_kwargs.get("leader_face_kwargs", {}),
                        **marqueeplot_outset_kwargs.get(
                            "leader_face_kwargs", {}
                        ),
                    },
                    "mark_glyph_kwargs": {
                        **marqueeplot_kwargs.get("mark_glyph_kwargs", {}),
                        **marqueeplot_outset_kwargs.get(
                            "mark_glyph_kwargs", {}
                        ),
                    },
                },
            )

        self._marqueeplot_outset = marqueeplot_outset

    def equalize_aspect(self: "OutsetGrid") -> "OutsetGrid":
        """Adjust axes {x,y}lims to ensure an equal xlim-to-ylim ratio across
        all axes.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        if self.source_axes is not None:
            aspect = calc_aspect(self.source_axes)
            for ax in self.axes.flat[1:]:
                set_aspect(ax, aspect)
        else:
            equalize_aspect(self.axes.flat)
        return self

    def marqueeplot(
        self: "OutsetGrid", equalize_aspect: bool = True
    ) -> "OutsetGrid":
        """Dispatch marquee annotation rendering for all subplots --- outset as
        well as source, if included.

        Parameters
        ----------
        equalize_aspect : bool, optional, default: True
            If True, adjusts axes limits to enforce equal ylim height to xlim
            width ratio.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        self.marqueeplot_source(equalize_aspect=False)
        self.marqueeplot_outset(equalize_aspect=False)
        if equalize_aspect:
            self.equalize_aspect()
        return self

    def marqueeplot_outset(
        self: "OutsetGrid", equalize_aspect: bool = True
    ) -> "OutsetGrid":
        """Dispatch marquee annotation rendering for outset plots only

        Parameters
        ----------
        equalize_aspect : bool, optional, default: True
            If True, adjusts axes limits to enforce equal ylim height to xlim
            width ratio.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        self._marqueeplot_outset(self)
        self._marqueeplot_outset = lambda self_: warnings.warn(
            "Redundant call to marqueeplot_outset, marquees were already drawn",
        )
        if equalize_aspect:
            self.equalize_aspect()
        return self

    def marqueeplot_source(
        self: "OutsetGrid", equalize_aspect: bool = True
    ) -> "OutsetGrid":
        """Dispatch marquee annotation rendering for the source plot only, if
        included.

        Parameters
        ----------
        equalize_aspect : bool, optional, default: True
            If True, adjusts axes limits to enforce equal ylim height to xlim
            width ratio.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        self._marqueeplot_source(self)
        self._marqueeplot_source = lambda self_: warnings.warn(
            "Redundant call to marqueeplot_source, marquees were already drawn",
        )
        if equalize_aspect:
            self.equalize_aspect()
        return self

    def map(self: "OutsetGrid") -> None:
        """Placeholder, raises NotImplementedError."""
        raise NotImplementedError()

    def map_outset(self: "OutsetGrid") -> None:
        """Placeholder, raises NotImplementedError."""
        raise NotImplementedError()

    def map_source(self: "OutsetGrid") -> None:
        """Placeholder, raises NotImplementedError."""
        raise NotImplementedError()

    def map_dataframe(
        self: "OutsetGrid", plotter: typing.Callable, *args, **kwargs
    ) -> "OutsetGrid":
        """Map a plotting function over all axes, including source plot axes
        (if present).

        Uses data stored at initialization, passed as a first `data=` kwarg. To
        plot other data, use `broadcast`. Complete dataset is used for source
        axes and faceted subsets are used for each outset axes.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        self.map_dataframe_outset(plotter, *args, **kwargs)
        self.map_dataframe_source(plotter, *args, **kwargs)
        return self

    def map_dataframe_outset(
        self: "OutsetGrid", plotter: typing.Callable, *args, **kwargs
    ) -> "OutsetGrid":
        """Map a plotting function over outset axes only.

        Uses data stored at initialization, passed as a first `data=` kwarg and
        faceted per subplot axes. To plot other data, use `broadcast_outset`.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        if "hue" in kwargs and self._hue_var is not None:
            raise ValueError("Cannot map `hue` if FacetGrid `hue` is set.")
        elif "hue" in kwargs and kwargs.get("hue_order", None) is None:
            assert self.hue_names is None
            kwargs["hue_order"] = sorted(self.__data[kwargs["hue"]].unique())
        super().map_dataframe(plotter, *args, **kwargs)
        return self

    def map_dataframe_source(
        self: "OutsetGrid", plotter: typing.Callable, *args, **kwargs
    ) -> "OutsetGrid":
        """Map a plotting function over the source plot axes only.

        If source plot axes are not enabled, performs no-op.

        Uses data stored at initialization, passed as a first `data=` kwarg.
        To plot other data, use `broadcast_source`.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        if self._hue_var is not None and "hue" in kwargs:
            raise ValueError("Cannot map `hue` if FacetGrid `hue` is set.")
        if "hue_order" in kwargs and not (
            "hue" in kwargs or self._hue_var is not None
        ):
            raise ValueError("Cannot map `hue_order` if `hue` is unset.")
        elif "hue_order" in kwargs and self.hue_names is None:
            raise ValueError(
                "Cannot map `hue_order` if FacetGrid `hue_order` is set.",
            )

        hue = opyt.or_value(self._hue_var, kwargs.pop("hue", None))
        hue_order = opyt.or_value(self.hue_names, kwargs.pop("hue_order", None))
        if hue is not None:
            kwargs["hue"] = hue
        if hue_order is not None:
            kwargs["hue_order"] = hue_order
        if self.source_axes is not None:
            plotter(self.__data, *args, ax=self.source_axes, **kwargs)
        self.tight_layout()
        return self

    def broadcast(
        self: "OutsetGrid",
        plotter: typing.Callable,
        *args,
        **kwargs,
    ) -> "OutsetGrid":
        """Map a plotting function over all axes, including the source plot
        axis (if present).

        Performs call with same data and arguments for all axes. To use faceted
        data stored at initialization, refer to `map_dataframe`.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.

        Notes
        -----
        Does not use data stored from initialization. Data should be provided
        via argument to this method.

        Preserves axis limits for all axes except the source plot, if present.
        """
        self.broadcast_outset(plotter, *args, **kwargs)
        self.broadcast_source(plotter, *args, **kwargs)
        return self

    def broadcast_outset(
        self: "OutsetGrid",
        plotter: typing.Callable,
        *args,
        **kwargs,
    ) -> "OutsetGrid":
        """Map a plotting function over only outset axes.

        Performs call with same data and arguments for all axes. To use faceted
        data stored at initialization, refer to `map_dataframe`.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.

        Notes
        -----
        Does not use data stored from initialization. Data should be provided
        via argument to this method.

        Preserves axis limits.
        """
        for ax in self.outset_axes:
            # store and restore axis limits, except for source plot if present
            xlim, ylim = ax.get_xlim(), ax.get_ylim()
            try:
                plotter(*args, ax=ax, **kwargs)
            except (TypeError, AttributeError):
                plt.sca(ax)
                plotter(*args, **kwargs)
            ax.set_xlim(*xlim)
            ax.set_ylim(*ylim)
        self.tight_layout()
        return self

    def broadcast_source(
        self: "OutsetGrid",
        plotter: typing.Callable,
        *args,
        **kwargs,
    ) -> "OutsetGrid":
        """Map a plotting function over source plot axes, if present.

        Performs call with same data and arguments for all axes. To use faceted
        data stored at initialization, refer to `map_dataframe`.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.

        Notes
        -----
        Does not use data stored from initialization. Data should be provided
        via argument to this method.

        Doesn't preserve axis limits.
        """
        if self.source_axes is not None:
            ax = self.source_axes
            try:
                plotter(*args, ax=ax, **kwargs)
            except (TypeError, AttributeError):
                plt.sca(ax)
                plotter(*args, **kwargs)
        self.tight_layout()
        return self



from matplotlib import markers as mpl_markers


def rotate_marker(marker: str, rotate_angle: float) -> mpl_markers.MarkerStyle:
    """Rotate a matplotlib marker by a specified angle."""
    marker_handle = mpl_markers.MarkerStyle(marker=marker)
    marker_handle._transform = marker_handle.get_transform().rotate_deg(
        rotate_angle,
    )
    return marker_handle



import numpy as np
import pytest

from outset._auxlib.compose_callout_leader_ import compose_callout_leader


# smoke test
@pytest.mark.parametrize(
    "rect_xlim, rect_ylim, ax_xlim, ax_ylim, stretch",
    [
        ((0, 1), (0, 1), (0, 5), (0, 5), 0.14),
        ((-1, 0), (-1, 0), (-5, 0), (-5, 0), 0.2),
        ((1, 4), (2, 3), (0, 5), (0, 5), 0.1),
    ],
)
def test_compose_callout_leader(
    rect_xlim, rect_ylim, ax_xlim, ax_ylim, stretch
):
    result = compose_callout_leader(
        rect_xlim, rect_ylim, ax_xlim, ax_ylim, stretch
    )
    assert len(result) == 4
    assert all(len(item) == 2 for item in result)

    _lowerleft, upperright = zip(rect_xlim, rect_ylim)
    upperleft, lowerright = zip(rect_xlim, reversed(rect_ylim))
    assert upperleft in result
    assert upperright in result
    assert lowerright in result

    for item in result:
        assert (
            np.array_equal(item, upperleft)
            or np.array_equal(item, lowerright)
            or np.array_equal(item, upperright)
            or (item[0] > rect_xlim[1] and item[1] > rect_ylim[1])
        )

    def test_invalid_limits():
        with pytest.raises(ValueError):
            compose_callout_leader((1, 0), (0, 2), (0, 5), (0, 5), 0.14)



import numbers
import typing

import frozendict
import numpy as np
from matplotlib import axes as mpl_axes
from matplotlib import patches as mpl_patches
from matplotlib import pyplot as plt
import seaborn as sns

from ._auxlib.draw_callout_ import draw_callout
from ._auxlib.draw_frame_ import draw_frame
from ._auxlib.is_axes_unset_ import is_axes_unset
from .MarkMagnifyingGlass_ import mark_magnifying_glass


def draw_marquee(
    frame_xlim: typing.Tuple[float, float],
    frame_ylim: typing.Tuple[float, float],
    ax: typing.Optional[mpl_axes.Axes] = None,
    *,
    color: typing.Optional[str] = "blue",
    clip_on: bool = False,
    despine: bool = True,
    frame_edge_kwargs: typing.Dict = frozendict.frozendict(),
    frame_face_kwargs: typing.Dict = frozendict.frozendict(),
    frame_inner_pad: typing.Union[float, typing.Tuple[float, float]] = 0.0,
    frame_outer_pad: typing.Union[float, typing.Tuple[float, float]] = 0.1,
    label: typing.Optional[str] = None,
    leader_edge_kwargs: typing.Dict = frozendict.frozendict(),
    leader_face_kwargs: typing.Dict = frozendict.frozendict(),
    leader_stretch: float = 0.1,
    mark_glyph: typing.Optional[typing.Callable] = mark_magnifying_glass,
    mark_glyph_kwargs: typing.Dict = frozendict.frozendict(),
    mark_retract: float = 0.1,
    zorder: float = 0,
) -> mpl_axes.Axes:
    """Mark a rectangular region on a matplotlib axes object, framing it with a
    zoom-effect callout.

    Consists of (1) a rectangular frame (2) a callout leader, and (3) a marked
    glyph. The rectangular frame consists of an outer of border and an
    underlaid solid color fill. The leader is a gradient fill clipped to angle
    to a point up and to to the right of the framed region. The marked glyph
    (e.g., a numeral, an asterisk, etc.) is drawn at the vertex point of the
    leader.

    The callout is capped by a customizable glyph, default as a magnifying
    glass.
        Parameters
        ----------
        frame_xlim : Tuple[float, float]
            X-limits (xmin, xmax) of the area to be marked as outset.
        frame_ylim : Tuple[float, float]
            Y-limits (ymin, ymax) of the area to be marked as outset.
        ax : matplotlib.axes.Axes, optional
            Axes object to draw the outset on. Defaults to `plt.gca()`.
        color : str, optional
            Color for the frame's edge and zoom indication lines.
        clip_on : bool, default False
            If True, drawing elements are clipped to the axes bounding box.
        despine : bool, default True
            If True, removes top and right spines from the plot.
        frame_edge_kwargs : Dict, default {}
            Customization arguments for the frame's edge.
        frame_face_kwargs : Dict, default {}
            Customization arguments for the frame's face.
        frame_inner_pad : Union[float, Tuple[float, float]], default 0.0
            Padding from data extent to frame boundary, calculated relative to data
            extent (float) or in absolute units (tuple).
        frame_outer_pad : Union[float, Tuple[float, float]], default 0.1
            Padding from frame boundary to axis viewport, calculated relative to data extent (float) or in absolute units (tuple).
        label : str, optional
            Label used for legend creation.
        leader_edge_kwargs : Dict, default {}
            Customization arguments for the leader's edge.
        leader_face_kwargs : Dict, default {}
            Customization arguments for the leader's face.
        leader_stretch : float, default 0.1
            Scale of callout leader relative to axis viewport.
        mark_glyph : Callable, optional
            A callable to draw a glyph at the outer vertex of the callout leader.

            If None, no glyph is drawn.
        mark_glyph_kwargs : Dict, default frozendict.frozendict()
            Arguments for the mark_glyph callable.
        mark_retract : float, default 0.1
            Fraction to pull back glyph from the outer vertex of the callout.
        zorder : float, default 0
            Z-order for layering plot elements.

        Returns
        -------
        matplotlib.axes.Axes
            Axes with the outset and annotations added.

        Notes
        -----
        Delegates to `_auxlib.draw_callout_.draw_callout` and
        `_auslib.draw_callout_.draw_frame` for drawing.

        See Also
        --------
        marqueeplot
            Figure-level interface for `draw_marquee`.
        OutsetGrid
            Figure-level interface for creating plots with marquee annotations.
    """
    if ax is None:
        ax = plt.gca()
    if color is None:
        color = sns.color_palette()[0]

    # pad frame coordinates out from data
    if isinstance(frame_inner_pad, tuple):
        pad_x, pad_y = frame_inner_pad
    elif isinstance(frame_inner_pad, numbers.Number):
        pad_x = (np.ptp(frame_xlim) or np.ptp(ax.get_xlim())) * frame_inner_pad
        pad_y = (np.ptp(frame_ylim) or np.ptp(ax.get_ylim())) * frame_inner_pad
    else:
        raise ValueError(
            f"frame_inner_pad must be float or tuple, not {frame_inner_pad}",
        )
    frame_xlim = np.array(frame_xlim) + np.array([-pad_x, pad_x])
    frame_ylim = np.array(frame_ylim) + np.array([-pad_y, pad_y])

    # pad axis viewport out from frame
    ax_xlim, ax_ylim = ax.get_xlim(), ax.get_ylim()
    if isinstance(frame_outer_pad, tuple):
        pad_x, pad_y = frame_outer_pad
    elif isinstance(frame_outer_pad, numbers.Number):
        pad_x = max(np.ptp(ax.get_xlim()), np.ptp(frame_xlim)) * frame_outer_pad
        pad_y = max(np.ptp(ax.get_ylim()), np.ptp(frame_ylim)) * frame_outer_pad
    else:
        raise ValueError(
            f"frame_outer_pad must be float or tuple, not {frame_outer_pad}",
        )
    if is_axes_unset(ax):  # ... axes are empty, so ignore current axis viewport
        if pad_x or np.ptp(frame_xlim):
            ax.set_xlim(frame_xlim[0] - pad_x, frame_xlim[1] + pad_x)
        if pad_y or np.ptp(frame_ylim):
            ax.set_ylim(frame_ylim[0] - pad_y, frame_ylim[1] + pad_y)
    else:  # if axes not empty or axlim already set, ensure no viewport shrink
        ax.set_xlim(
            min(frame_xlim[0] - pad_x, ax_xlim[0]),
            max(frame_xlim[1] + pad_x, ax_xlim[1]),
        )
        ax.set_ylim(
            min(frame_ylim[0] - pad_y, ax_ylim[0]),
            max(frame_ylim[1] + pad_y, ax_ylim[1]),
        )

    # tweak zorder to ensure multiple outset annotations layer properly
    ax_width, ax_height = np.ptp(ax.get_xlim()), np.ptp(ax.get_ylim())
    ax_diag = np.sqrt(ax_width**2 + ax_height**2)
    upper_right_drop_x = ax.get_ylim()[1] - frame_xlim[1]
    upper_right_drop_y = ax.get_ylim()[1] - frame_ylim[1]
    upper_right_drop = np.sqrt(
        upper_right_drop_x**2 + upper_right_drop_y**2
    )
    zorder += 0.25 * upper_right_drop / ax_diag

    # Frame outset region
    ###########################################################################
    draw_frame(
        frame_xlim,
        frame_ylim,
        ax=ax,
        clip_on=clip_on,
        frame_edge_kwargs=frame_edge_kwargs,
        frame_face_kwargs=frame_face_kwargs,
        edgecolor=color,
        facecolor=color,
        zorder=zorder,
    )

    # Draw callout
    ###########################################################################
    draw_callout(
        frame_xlim,
        frame_ylim,
        ax,
        color=color,
        clip_on=clip_on,
        leader_edge_kwargs=leader_edge_kwargs,
        leader_face_kwargs=leader_face_kwargs,
        mark_glyph=mark_glyph,
        mark_glyph_kwargs=mark_glyph_kwargs,
        mark_retract=mark_retract,
        leader_stretch=leader_stretch,
        zorder=zorder,
    )

    # Finalize
    ###########################################################################
    ax.set_axisbelow(True)  # ensure annotations above if outside bounds
    if despine:
        ax.spines[["right", "top"]].set_visible(False)
    if label is not None:
        ax.legend(handles=[mpl_patches.Patch(color=color, label=label)])

    return ax



import matplotlib.pyplot as plt
import seaborn as sns

from outset import stub_all_clipped_values


def test_stub_all_clipped_values_up():
    ax = plt.gca()  # Get the current axes=
    sns.scatterplot(x=[0.5], y=[3], ax=ax)

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    # Call the function with test values
    stub_all_clipped_values(ax)

    # Save and print the path of the output image
    outpath = "/tmp/test_stub_all_clipped_values_up.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_stub_all_clipped_values_down():
    ax = plt.gca()  # Get the current axes=
    sns.scatterplot(x=[0.5], y=[-3], ax=ax)

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    # Call the function with test values
    stub_all_clipped_values(ax)

    # Save and print the path of the output image
    outpath = "/tmp/test_stub_all_clipped_values_down.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_stub_all_clipped_values_left():
    ax = plt.gca()  # Get the current axes=
    sns.scatterplot(y=[0.5], x=[-3], ax=ax)

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    # Call the function with test values
    stub_all_clipped_values(ax)

    # Save and print the path of the output image
    outpath = "/tmp/test_stub_all_clipped_values_left.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_stub_all_clipped_values_right():
    ax = plt.gca()  # Get the current axes
    sns.scatterplot(y=[0.5], x=[3], ax=ax)

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    # Call the function with test values
    stub_all_clipped_values(ax)

    # Save and print the path of the output image
    outpath = "/tmp/test_stub_all_clipped_values_right.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_stub_all_clipped_values_upright():
    ax = plt.gca()  # Get the current axes=
    sns.scatterplot(x=[2], y=[3], ax=ax)

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    # Call the function with test values
    stub_all_clipped_values(ax)

    # Save and print the path of the output image
    outpath = "/tmp/test_stub_all_clipped_values_upright.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_stub_all_clipped_values_with_error_bar():
    ax = plt.gca()  # Get the current axes

    # Plot a point with an error bar
    x, y = [2.5], [0.5]
    x_err, y_err = [0.01], [0.02]  # Large vertical error bar
    sns.scatterplot(x=x, y=y, ax=ax)
    plt.errorbar(x, y, xerr=x_err, yerr=y_err, fmt="o", color="blue")

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    # Call the function with test values
    stub_all_clipped_values(ax)

    # Save and print the path of the output image
    outpath = "/tmp/test_stub_all_clipped_values_with_error_bar.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")
